@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2007-2011 Edgewall Software, 2013-2018 the Babel team
+# Copyright (C) 2007-2011 Edgewall Software, 2013-2020 the Babel team
 # All rights reserved.
 #
 # This software is licensed as described in the file LICENSE, which
@@ -16,11 +16,10 @@ import pytest
 
 from datetime import date
 
-from babel import Locale, localedata, numbers
+from babel import localedata, numbers
 from babel.numbers import (
     list_currencies, validate_currency, UnknownCurrencyError, is_currency, normalize_currency,
-    get_currency_precision, get_decimal_precision)
-from babel.localedata import locale_identifiers
+    get_currency_precision, get_decimal_precision, get_currency_unit_pattern)
 from babel._compat import decimal
 
 
@@ -154,6 +153,36 @@ class FormatDecimalTestCase(unittest.TestCase):
         fmt = numbers.format_decimal(number, format="@@@", locale='en_US')
         self.assertEqual('0.000000700', fmt)
 
+    def test_group_separator(self):
+        self.assertEqual('29567.12', numbers.format_decimal(29567.12,
+                                                                     locale='en_US', group_separator=False))
+        self.assertEqual('29567,12', numbers.format_decimal(29567.12,
+                                                                     locale='fr_CA', group_separator=False))
+        self.assertEqual('29567,12', numbers.format_decimal(29567.12,
+                                                                     locale='pt_BR', group_separator=False))
+        self.assertEqual(u'$1099.98', numbers.format_currency(1099.98, 'USD',
+                                                             locale='en_US', group_separator=False))
+        self.assertEqual(u'101299,98\xa0€', numbers.format_currency(101299.98, 'EUR',
+                                                            locale='fr_CA', group_separator=False))
+        self.assertEqual('101299.98 euros', numbers.format_currency(101299.98, 'EUR',
+                                                            locale='en_US', group_separator=False, format_type='name'))
+        self.assertEqual(u'25123412\xa0%', numbers.format_percent(251234.1234, locale='sv_SE', group_separator=False))
+
+        self.assertEqual(u'29,567.12', numbers.format_decimal(29567.12,
+                                                            locale='en_US', group_separator=True))
+        self.assertEqual(u'29\u202f567,12', numbers.format_decimal(29567.12,
+                                                            locale='fr_CA', group_separator=True))
+        self.assertEqual(u'29.567,12', numbers.format_decimal(29567.12,
+                                                            locale='pt_BR', group_separator=True))
+        self.assertEqual(u'$1,099.98', numbers.format_currency(1099.98, 'USD',
+                                                              locale='en_US', group_separator=True))
+        self.assertEqual(u'101\u202f299,98\xa0\u20ac', numbers.format_currency(101299.98, 'EUR',
+                                                                    locale='fr_CA', group_separator=True))
+        self.assertEqual(u'101,299.98 euros', numbers.format_currency(101299.98, 'EUR',
+                                                                    locale='en_US', group_separator=True,
+                                                                    format_type='name'))
+        self.assertEqual(u'25\xa0123\xa0412\xa0%', numbers.format_percent(251234.1234, locale='sv_SE', group_separator=True))
+
 
 class NumberParsingTestCase(unittest.TestCase):
 
@@ -165,6 +194,36 @@ class NumberParsingTestCase(unittest.TestCase):
         self.assertRaises(numbers.NumberFormatError,
                           lambda: numbers.parse_decimal('2,109,998', locale='de'))
 
+    def test_parse_decimal_strict_mode(self):
+        # Numbers with a misplaced grouping symbol should be rejected
+        with self.assertRaises(numbers.NumberFormatError) as info:
+            numbers.parse_decimal('11.11', locale='de', strict=True)
+        assert info.exception.suggestions == ['1.111', '11,11']
+        # Numbers with two misplaced grouping symbols should be rejected
+        with self.assertRaises(numbers.NumberFormatError) as info:
+            numbers.parse_decimal('80.00.00', locale='de', strict=True)
+        assert info.exception.suggestions == ['800.000']
+        # Partially grouped numbers should be rejected
+        with self.assertRaises(numbers.NumberFormatError) as info:
+            numbers.parse_decimal('2000,000', locale='en_US', strict=True)
+        assert info.exception.suggestions == ['2,000,000', '2,000']
+        # Numbers with duplicate grouping symbols should be rejected
+        with self.assertRaises(numbers.NumberFormatError) as info:
+            numbers.parse_decimal('0,,000', locale='en_US', strict=True)
+        assert info.exception.suggestions == ['0']
+        # Return only suggestion for 0 on strict
+        with self.assertRaises(numbers.NumberFormatError) as info:
+            numbers.parse_decimal('0.00', locale='de', strict=True)
+        assert info.exception.suggestions == ['0']
+        # Properly formatted numbers should be accepted
+        assert str(numbers.parse_decimal('1.001', locale='de', strict=True)) == '1001'
+        # Trailing zeroes should be accepted
+        assert str(numbers.parse_decimal('3.00', locale='en_US', strict=True)) == '3.00'
+        # Numbers without any grouping symbol should be accepted
+        assert str(numbers.parse_decimal('2000.1', locale='en_US', strict=True)) == '2000.1'
+        # High precision numbers should be accepted
+        assert str(numbers.parse_decimal('5,000001', locale='fr', strict=True)) == '5.000001'
+
 
 def test_list_currencies():
     assert isinstance(list_currencies(), set)
@@ -179,7 +238,7 @@ def test_list_currencies():
 
     assert list_currencies(locale='pa_Arab') == {'PKR', 'INR', 'EUR'}
 
-    assert len(list_currencies()) == 300
+    assert len(list_currencies()) == 303
 
 
 def test_validate_currency():
@@ -228,6 +287,17 @@ def test_get_currency_precision():
     assert get_currency_precision('JPY') == 0
 
 
+def test_get_currency_unit_pattern():
+    assert get_currency_unit_pattern('USD', locale='en_US') == '{0} {1}'
+    assert get_currency_unit_pattern('USD', locale='es_GT') == '{1} {0}'
+
+    # 'ro' locale various pattern according to count
+    assert get_currency_unit_pattern('USD', locale='ro', count=1) == '{0} {1}'
+    assert get_currency_unit_pattern('USD', locale='ro', count=2) == '{0} {1}'
+    assert get_currency_unit_pattern('USD', locale='ro', count=100) == '{0} de {1}'
+    assert get_currency_unit_pattern('USD', locale='ro') == '{0} de {1}'
+
+
 def test_get_territory_currencies():
     assert numbers.get_territory_currencies('AT', date(1995, 1, 1)) == ['ATS']
     assert numbers.get_territory_currencies('AT', date(2011, 1, 1)) == ['EUR']
@@ -371,7 +441,7 @@ def test_format_currency_format_type():
     assert (numbers.format_currency(1099.98, 'JPY', locale='en_US')
             == u'\xa51,100')
     assert (numbers.format_currency(1099.98, 'COP', u'#,##0.00', locale='es_ES')
-            == u'1.100')
+            == u'1.099,98')
     assert (numbers.format_currency(1099.98, 'JPY', locale='en_US',
                                     currency_digits=False)
             == u'\xa51,099.98')
@@ -415,6 +485,52 @@ def test_format_currency_quantization():
             '0.9999999999', 'USD', locale=locale_code, decimal_quantization=False).find('9999999999') > -1
 
 
+def test_format_currency_long_display_name():
+    assert (numbers.format_currency(1099.98, 'USD', locale='en_US', format_type='name')
+            == u'1,099.98 US dollars')
+    assert (numbers.format_currency(1.00, 'USD', locale='en_US', format_type='name')
+            == u'1.00 US dollar')
+    assert (numbers.format_currency(1.00, 'EUR', locale='en_US', format_type='name')
+            == u'1.00 euro')
+    assert (numbers.format_currency(2, 'EUR', locale='en_US', format_type='name')
+            == u'2.00 euros')
+    # This tests that '{1} {0}' unitPatterns are found:
+    assert (numbers.format_currency(1, 'USD', locale='sw', format_type='name')
+            == u'dola ya Marekani 1.00')
+    # This tests unicode chars:
+    assert (numbers.format_currency(1099.98, 'USD', locale='es_GT', format_type='name')
+            == u'dólares estadounidenses 1,099.98')
+    # Test for completely unknown currency, should fallback to currency code
+    assert (numbers.format_currency(1099.98, 'XAB', locale='en_US', format_type='name')
+            == u'1,099.98 XAB')
+
+    # Test for finding different unit patterns depending on count
+    assert (numbers.format_currency(1, 'USD', locale='ro', format_type='name')
+            == u'1,00 dolar american')
+    assert (numbers.format_currency(2, 'USD', locale='ro', format_type='name')
+            == u'2,00 dolari americani')
+    assert (numbers.format_currency(100, 'USD', locale='ro', format_type='name')
+            == u'100,00 de dolari americani')
+
+
+def test_format_currency_long_display_name_all():
+    for locale_code in localedata.locale_identifiers():
+        assert numbers.format_currency(
+            1, 'USD', locale=locale_code, format_type='name').find('1') > -1
+        assert numbers.format_currency(
+            '1', 'USD', locale=locale_code, format_type='name').find('1') > -1
+
+
+def test_format_currency_long_display_name_custom_format():
+    assert (numbers.format_currency(1099.98, 'USD', locale='en_US',
+                                    format_type='name', format='##0')
+            == '1099.98 US dollars')
+    assert (numbers.format_currency(1099.98, 'USD', locale='en_US',
+                                    format_type='name', format='##0',
+                                    currency_digits=False)
+            == '1100 US dollars')
+
+
 def test_format_percent():
     assert numbers.format_percent(0.34, locale='en_US') == u'34%'
     assert numbers.format_percent(0, locale='en_US') == u'0%'
@@ -593,3 +709,18 @@ def test_numberpattern_repr():
 def test_parse_static_pattern():
     assert numbers.parse_pattern('Kun')  # in the So locale in CLDR 30
     # TODO: static patterns might not be correctly `apply()`ed at present
+
+
+def test_parse_decimal_nbsp_heuristics():
+    # Re https://github.com/python-babel/babel/issues/637 –
+    #    for locales (of which there are many) that use U+00A0 as the group
+    #    separator in numbers, it's reasonable to assume that input strings
+    #    with plain spaces actually should have U+00A0s instead.
+    #    This heuristic is only applied when strict=False.
+    n = decimal.Decimal("12345.123")
+    assert numbers.parse_decimal("12 345.123", locale="fi") == n
+    assert numbers.parse_decimal(numbers.format_decimal(n, locale="fi"), locale="fi") == n
+
+
+def test_very_small_decimal_no_quantization():
+    assert numbers.format_decimal(decimal.Decimal('1E-7'), locale='en', decimal_quantization=False) == '0.0000001'
